#!/usr/bin/env python
"""
This script compares two benchmark run dir and analyzes the changes in test performance.
It categorizes tests as improved, worsened, stable, or present in only one of the benchmark runs.
"""
from dataclasses import dataclass
import os
import subprocess
from collections import Counter
from functools import total_ordering
from importlib.resources.abc import Traversable
from pathlib import Path
from typing import NamedTuple, Union

from datetime import timedelta
import sys
from dotenv import load_dotenv

if __name__ == "__main__":
    sys.path.append(str(Path(__file__).parent.parent))
from .benchmark_analyzer import BenchmarkAnalyzer


def _get_visual_indicator(percent_change: float | None) -> str:
    """Generate a visual indicator string based on percentage change."""
    if percent_change is None:
        return ""
    if percent_change == 0:
        return ""
    # Convert to absolute value to determine length, but keep sign for direction
    abs_change = abs(percent_change)
    indicator_length = min(20, max(1, int(abs_change / 10)))  # 1 char per 10% change, max 20
    return " " + ("+" if percent_change > 0 else "-") * indicator_length

def _get_token_change_indicators(test_1: 'AiderTestResult', test_2: 'AiderTestResult') -> tuple[str, str]:
    """Generate visual indicators for token changes between two test runs.

    Returns:
        tuple containing:
            - Left-aligned received tokens indicator in parentheses
            - Right-aligned sent tokens indicator in parentheses
    """
    sent_change = ((test_2.sent_tokens - test_1.sent_tokens) * 100 / test_1.sent_tokens) if test_1.sent_tokens else 0
    recv_change = ((test_2.received_tokens - test_1.received_tokens) * 100 / test_1.received_tokens) if test_1.received_tokens else 0

    # Generate the indicators (about 5% change per character, so 20 chars = ~100% change)
    sent_indicator = "+" * min(20, max(1, int(abs(sent_change) / 5)))
    recv_indicator = "-" * min(20, max(1, int(abs(recv_change) / 5)))

    # Right-align received tokens indicator (pad left with spaces)
    recv_col = f"({recv_indicator:>20})"

    # Left-align sent tokens indicator (pad right with spaces)
    sent_col = f"({sent_indicator:<20})"

    return recv_col, sent_col  # received tokens first, then sent tokens


@dataclass
class StatusPrinter:
    test_count: int

    def __call__(self, prefix: str, count: int, indent: str = "", suffix: str = "") -> None:
        """Print a status line with consistent formatting."""
        if count == 0:
            return
        current_percent = count * 100 / self.test_count
        print(f"{indent}{prefix}: {count:3d} ({current_percent:3.0f}%){_get_visual_indicator(current_percent)}{suffix}")


def main(run1path: Traversable, run2path: Traversable):
    """
    Main function to compare two benchmark runs and print the analysis.

    Args:
    benchmark1 (Traversable): Path to the first benchmark run.
    benchmark2 (Traversable): Path to the second benchmark run.

    This function parses both benchmark runs, compares them, and prints a detailed analysis
    of how tests have changed between the two runs. It categorizes tests as improved,
    worsened, stable, or present in only one run, and provides a summary count for each category and sub-category.
    """
    print(f"--- {run1path.name}")
    print(f"+++ {run2path.name}")
    print("# ============= Failed Attempts per Test =============")
    print("# N >= 0: It eventually passed after N failed attempts")
    print("# N < 0 : All attempts failed and the limit was reached")
    benchmark_run_1 = {t.name: t for t in parse_benchmark_source(run1path)}
    benchmark_run_2 = {t.name: t for t in parse_benchmark_source(run2path)}

    (
        test_names_only_1, test_names_only_2, test_names_improved, test_names_worsened, test_names_stable
    ) = compare_benchmark_runs(benchmark_run_1, benchmark_run_2)

    test_names_only_1_passed = [t for t in test_names_only_1 if benchmark_run_1[t].failed_attempt_count >= 0]
    if test_names_only_1_passed:
        print()
        print(f"@@ REMOVED ({len(test_names_only_1_passed)} PASSED) @@")
        for test_name in test_names_only_1_passed:
            failed_attempt_count = benchmark_run_1[test_name].failed_attempt_count
            print(f"<{'-' if failed_attempt_count < 0 else '+'}{test_name}: {failed_attempt_count}")

    test_names_only_1_failed = [t for t in test_names_only_1 if benchmark_run_1[t].failed_attempt_count < 0]
    if test_names_only_1_failed:
        print()
        print(f"@@ REMOVED ({len(test_names_only_1_failed)} FAILED) @@")
        for test_name in test_names_only_1_failed:
            failed_attempt_count = benchmark_run_1[test_name].failed_attempt_count
            print(f"<{'-' if failed_attempt_count < 0 else '+'}{test_name}: {failed_attempt_count}")

    test_names_only_2_passed = [t for t in test_names_only_2 if benchmark_run_2[t].failed_attempt_count >= 0]
    if test_names_only_2_passed:
        print()
        print(f"@@ NEW ({len(test_names_only_2_passed)} PASSED) @@")
        for test_name in test_names_only_2_passed:
            failed_attempt_count = benchmark_run_2[test_name].failed_attempt_count
            print(f">{'-' if failed_attempt_count < 0 else '+'}{test_name}: {failed_attempt_count}")

    test_names_only_2_failed = [t for t in test_names_only_2 if benchmark_run_2[t].failed_attempt_count < 0]
    if test_names_only_2_failed:
        print()
        print(f"@@ NEW ({len(test_names_only_2_failed)} FAILED) @@")
        for test_name in test_names_only_2_failed:
            failed_attempt_count = benchmark_run_2[test_name].failed_attempt_count
            print(f">{'-' if failed_attempt_count < 0 else '+'}{test_name}: {failed_attempt_count}")

    test_names_improved_now_passes = [t for t in test_names_improved if benchmark_run_1[t].failed_attempt_count < 0]
    now_passed_count = len(test_names_improved_now_passes)
    if test_names_improved_now_passes:
        print()
        print(f"@@ Improved, now PASSED ({now_passed_count}) @@")
        print_test_line('++', test_names_improved_now_passes, benchmark_run_1, benchmark_run_2)

    test_names_improved_minor = [t for t in test_names_improved if benchmark_run_1[t].failed_attempt_count >= 0]
    if test_names_improved_minor:
        print()
        print(f"@@ Improved, minor ({len(test_names_improved_minor)}) @@")
        print_test_line('+', test_names_improved_minor, benchmark_run_1, benchmark_run_2)

    test_names_worsened_now_fails = [t for t in test_names_worsened if benchmark_run_2[t].failed_attempt_count < 0]
    now_failed_count = len(test_names_worsened_now_fails)
    if test_names_worsened_now_fails:
        print()
        print(f"@@ Worsened, now FAILED ({now_failed_count}) @@")
        print_test_line('--', test_names_worsened_now_fails, benchmark_run_1, benchmark_run_2)

    test_names_worsened_minor = [t for t in test_names_worsened if benchmark_run_2[t].failed_attempt_count >= 0]
    if test_names_worsened_minor:
        print()
        print(f"@@ Worsened, still PASSED ({len(test_names_worsened_minor)}) @@")
        print_test_line('-', test_names_worsened_minor, benchmark_run_1, benchmark_run_2)

    test_names_stable_passed = [t for t in test_names_stable if benchmark_run_1[t].failed_attempt_count >= 0]
    if test_names_stable_passed:
        print()
        print(f"@@ Stable: PASSED ({len(test_names_stable_passed)}) @@")
        print_test_line('=+', test_names_stable_passed, benchmark_run_1, benchmark_run_2)

    test_names_stable_failed = [t for t in test_names_stable if benchmark_run_1[t].failed_attempt_count < 0]
    if test_names_stable_failed:
        print()
        print(f"@@ Stable: FAILED ({len(test_names_stable_failed)}) @@")
        print_test_line('=-', test_names_stable_failed, benchmark_run_1, benchmark_run_2)

    print()
    print(f"--- {run1path.name}")
    print(f"+++ {run2path.name}")
    printer = StatusPrinter(len(benchmark_run_2))
    printer(
        "@@ NPNF-Delta ",
        now_passed_count - now_failed_count,
        suffix=" @@"
    )
    if now_passed_count - now_failed_count == 0:
        print("@@ NPNF-Delta :   0 ( Both are equivalent ) @@")
    print("@@ ============= TEST STATUS CHANGES ============ @@")
    if test_names_only_1:
        printer("< REMOVED     ", len(test_names_only_1))
        printer("< +       PASS", len(test_names_only_1_passed))
        printer("< -       FAIL", len(test_names_only_1_failed))
    if test_names_only_2:
        printer("> NEW         ", len(test_names_only_2))
        printer("> +       PASS", len(test_names_only_2_passed))
        printer("> -       FAIL", len(test_names_only_2_failed))
    if test_names_stable:
        printer("# STABLE      ", len(test_names_stable))
        printer("#+        PASS", len(test_names_stable_passed))
        printer("#-        FAIL", len(test_names_stable_failed))
    if test_names_improved:
        printer("+ IMPROVED    ", len(test_names_improved))
        printer("++    Now PASS", now_passed_count)
        printer("+        Minor", len(test_names_improved_minor))
    if test_names_worsened:
        printer("- WORSENED    ", len(test_names_worsened))
        printer("--    Now FAIL", now_failed_count)
        printer("-        Minor", len(test_names_worsened_minor))

    test_count_delta = len(benchmark_run_2) - len(benchmark_run_1)
    # Calculate totals for each run
    sent_tokens_1 = sum(t.sent_tokens for t in benchmark_run_1.values())
    received_tokens_1 = sum(t.received_tokens for t in benchmark_run_1.values())
    duration_1 = sum(t.duration for t in benchmark_run_1.values())
    sent_tokens_2 = sum(t.sent_tokens for t in benchmark_run_2.values())
    received_tokens_2 = sum(t.received_tokens for t in benchmark_run_2.values())
    cost_1 = sum(t.cost for t in benchmark_run_1.values())
    cost_2 = sum(t.cost for t in benchmark_run_2.values())
    timeouts_1 = sum(t.timeouts for t in benchmark_run_1.values())
    timeouts_2 = sum(t.timeouts for t in benchmark_run_2.values())
    error_outputs_1 = sum(t.error_output_count for t in benchmark_run_1.values())
    error_outputs_2 = sum(t.error_output_count for t in benchmark_run_2.values())
    user_asks_1 = sum(t.user_ask_count for t in benchmark_run_1.values())
    user_asks_2 = sum(t.user_ask_count for t in benchmark_run_2.values())
    context_exhausts_1 = sum(t.exhausted_context_window_count for t in benchmark_run_1.values())
    context_exhausts_2 = sum(t.exhausted_context_window_count for t in benchmark_run_2.values())
    malformed_1 = sum(t.malformed_responses for t in benchmark_run_1.values())
    malformed_2 = sum(t.malformed_responses for t in benchmark_run_2.values())
    syntax_errors_1 = sum(t.syntax_errors for t in benchmark_run_1.values())
    syntax_errors_2 = sum(t.syntax_errors for t in benchmark_run_2.values())
    indent_errors_1 = sum(t.indentation_errors for t in benchmark_run_1.values())
    indent_errors_2 = sum(t.indentation_errors for t in benchmark_run_2.values())
    lazy_comments_1 = sum(t.lazy_comments for t in benchmark_run_1.values())
    lazy_comments_2 = sum(t.lazy_comments for t in benchmark_run_2.values())
    cedarscript_errors_1 = sum(t.cedarscript_errors for t in benchmark_run_1.values())
    cedarscript_errors_2 = sum(t.cedarscript_errors for t in benchmark_run_2.values())
    duration_2 = sum(t.duration for t in benchmark_run_2.values())

    max_failed_attempt_1, attempt_counts_1 = _get_attempt_limit_and_normalized_counts(benchmark_run_1)
    max_failed_attempt_2, attempt_counts_2 = _get_attempt_limit_and_normalized_counts(benchmark_run_2)

    print()
    print("@@ ============ Success Distribution =========== @@")
    all_attempt_counts = sorted(set(attempt_counts_1.keys()) | set(attempt_counts_2.keys()))
    for attempt_count in all_attempt_counts:
        count_1 = attempt_counts_1.get(attempt_count, 0)
        count_2 = attempt_counts_2.get(attempt_count, 0)
        if count_1 == 0 and count_2 == 0:
            continue
        count_diff = count_2 - count_1
        percent_change = (count_diff * 100 / count_1) if count_1 else None
        if attempt_count < 0:
            prefix = f"#              FAIL"
        else:
            prefix = f"#          Pass {attempt_count+1:3d}"

        print(f"{prefix}: {count_2:10d} {f'({count_diff:+10d}, {percent_change:+4.0f}%){_get_visual_indicator(percent_change)}' if count_1 else 'N/A'}")

    # TODO Print model and edit_format for each run (just take the first value)
    # Get model and edit format from first test in each run (they should be the same for all tests in a run)
    model_1 = next(iter(benchmark_run_1.values())).model if benchmark_run_1 else "N/A"
    model_2 = next(iter(benchmark_run_2.values())).model if benchmark_run_2 else "N/A"
    edit_format_1 = next(iter(benchmark_run_1.values())).edit_format if benchmark_run_1 else "N/A"
    edit_format_2 = next(iter(benchmark_run_2.values())).edit_format if benchmark_run_2 else "N/A"
    print("@@ ================== METRICS  ================= @@")
    print(f"# MODEL            : {model_2.split('/')[-1]:>10} {'(was ' + model_1.split('/')[-1] + ')' if model_1 != model_2 else ''}")
    print(f"# EDIT FORMAT      : {edit_format_2:>10} {'(was ' + edit_format_1 + ')' if edit_format_1 != edit_format_2 else ''}")
    print(f"# TOTAL TEST COUNT : {len(benchmark_run_2):10d} {f'({test_count_delta:+10d}, {test_count_delta*100/len(benchmark_run_1):+4.0f}%){_get_visual_indicator(test_count_delta*100/len(benchmark_run_1) if test_count_delta else None)}' if test_count_delta else ''}")
    print(f"# Max attempt count: {max_failed_attempt_2:10d}{f" ({max_failed_attempt_2 - max_failed_attempt_1:+d})" if max_failed_attempt_2 != max_failed_attempt_1 else ""}")
    print(f"# DURATION hh:mm:ss:    {str(timedelta(seconds=int(duration_2)))} ({'-' if duration_2 < duration_1 else '+'}  {str(timedelta(seconds=int(abs(duration_2 - duration_1))))}, {(duration_2 - duration_1)*100/duration_1:+4.0f}%){_get_visual_indicator((duration_2 - duration_1)*100/duration_1)}")
    print(f"# COST ($)         : {cost_2:10,.2f} {f'({cost_2 - cost_1:+10,.2f}, {(cost_2 - cost_1)*100/cost_1:+4.0f}%){_get_visual_indicator((cost_2 - cost_1)*100/cost_1)}' if cost_1 else 'N/A'}")
    print(f"# TOKENS SENT      : {sent_tokens_2:10,} ({sent_tokens_2 - sent_tokens_1:+10,}, {(sent_tokens_2 - sent_tokens_1)*100/sent_tokens_1:+4.0f}%){_get_visual_indicator((sent_tokens_2 - sent_tokens_1)*100/sent_tokens_1)}")
    print(f"# TOKENS RECEIVED  : {received_tokens_2:10,} ({received_tokens_2 - received_tokens_1:+10,}, {(received_tokens_2 - received_tokens_1)*100/received_tokens_1:+4.0f}%){_get_visual_indicator((received_tokens_2 - received_tokens_1)*100/received_tokens_1)}")
    print_metric_diff("CEDARSCRIPT ERR. ", cedarscript_errors_1, cedarscript_errors_2)
    print_metric_diff("ERROR OUTPUTS    ", error_outputs_1, error_outputs_2)
    print_metric_diff("MALFORMED        ", malformed_1, malformed_2)
    print_metric_diff("SYNTAX ERRORS    ", syntax_errors_1, syntax_errors_2)
    print_metric_diff("INDENT ERRORS    ", indent_errors_1, indent_errors_2)
    print_metric_diff("LAZY COMMENTS    ", lazy_comments_1, lazy_comments_2)
    print_metric_diff("USER ASKS        ", user_asks_1, user_asks_2)
    print_metric_diff("CONTEXT EXHAUSTS ", context_exhausts_1, context_exhausts_2)
    print_metric_diff("TIMEOUTS         ", timeouts_1, timeouts_2)


def print_test_line(prefix: str, tests_names, benchmark_run_1, benchmark_run_2):
    for test_name in tests_names:
        failed_attempts_1 = benchmark_run_1[test_name].failed_attempt_count
        failed_attempts_2 = benchmark_run_2[test_name].failed_attempt_count
        failed_attempt_delta = "          " if failed_attempts_1 == 0 and failed_attempts_2 == 0 else f"[{failed_attempts_1:2} -> {failed_attempts_2:2}]"
        sent_ind, recv_ind = _get_token_change_indicators(benchmark_run_1[test_name], benchmark_run_2[test_name])
        word_count = benchmark_run_1[test_name].word_count
        cedarscript_errors_1 = benchmark_run_1[test_name].cedarscript_errors
        cedarscript_errors_2 = benchmark_run_2[test_name].cedarscript_errors
        cedarscript_delta = cedarscript_errors_2 - cedarscript_errors_1
        cedarscript_delta = f"[{cedarscript_delta:+3d}]" if cedarscript_delta else "     "
        print(
            f"{prefix} {failed_attempt_delta} {sent_ind} {recv_ind} "
            f"{cedarscript_delta} {word_count:5d} {test_name}")


def print_metric_diff(metric_name, value_run_1, value_run_2):
    print(
        f"# {metric_name}: {value_run_2:10d} {f"({value_run_2 - value_run_1:+10d}, {(value_run_2 - value_run_1) * 100 / value_run_1:+4.0f}%){_get_visual_indicator((value_run_2 - value_run_1) * 100 / value_run_1 if value_run_1 else None)}" if value_run_1 else 'N/A'}")


@total_ordering
class AiderTestResult(NamedTuple):
    """
    Represents the result of a benchmark test that may require multiple attempts to pass.

    This class tracks various metrics about test execution and provides normalized comparison
    of test results based on the number of failed attempts. The comparison logic treats all
    negative failed_attempt_count values (representing tests that hit max attempts) as equal,
    and considers them worse than any successful test (zero or positive failed_attempt_count).
    """

    failed_attempt_count: int
    word_count: int = 0
    name: str = ''
    duration: float = 0
    sent_tokens: int = 0
    received_tokens: int = 0
    model: str = ''
    edit_format: str = ''
    cost: float = 0
    timeouts: int = 0
    error_output_count: int = 0
    user_ask_count: int = 0
    exhausted_context_window_count: int = 0
    malformed_responses: int = 0
    syntax_errors: int = 0
    indentation_errors: int = 0
    lazy_comments: int = 0
    cedarscript_errors: int = 0

    def __int__(self) -> int:
        """
        Returns the raw failed_attempt_count value.

        A positive value indicates the number of failed attempts before success.
        Zero indicates success on first try.
        A negative value indicates the test never succeeded (hit max attempts).
        """
        return self.failed_attempt_count

    @property
    def norm_failed_attempts(self) -> int:
        """
        Returns a normalized version of failed_attempt_count for comparison purposes.

        All negative values are normalized to -1 since they represent the same state
        (test never passed within max attempts). Zero and positive values remain unchanged.

        Returns:
            int: -1 if failed_attempt_count is negative, otherwise the actual failed_attempt_count
        """
        return self._normalize_attempts(self.failed_attempt_count)

    @staticmethod
    def _normalize_attempts(value: int) -> int:
        """
        Normalizes the failed attempts count by converting all negative values to -1.

        Args:
            value: The raw failed attempts count

        Returns:
            int: -1 if value is negative, otherwise the original value
        """
        return -1 if value < 0 else value

    def __eq__(self, other: Union['AiderTestResult', int]) -> bool:
        """
        Compares this result with another result or an integer.

        Uses normalized failed attempts count where all negative values are treated as equal.
        Can compare against either another AiderTestResult instance or an integer.

        Args:
            other: Another AiderTestResult instance or an integer to compare with

        Returns:
            bool: True if the normalized failed attempts are equal, False otherwise
        """
        if isinstance(other, (int, AiderTestResult)):
            other_val = other.norm_failed_attempts if isinstance(other, AiderTestResult) else self._normalize_attempts(other)
            return self.norm_failed_attempts == other_val
        return False

    def __lt__(self, other: Union['AiderTestResult', int]) -> bool:
        """
        Determines if this result is "less than" another result or integer.

        Uses normalized failed attempts count where all negative values are treated as equal.
        A result with fewer failed attempts is considered "less than" one with more attempts.
        Any successful test (>= 0) is considered "less than" any failed test (< 0).

        Args:
            other: Another AiderTestResult instance or an integer to compare with

        Returns:
            bool: True if this result should be considered better than the other, False otherwise
        """
        if not isinstance(other, (int, AiderTestResult)):
            return NotImplemented

        other_val = other.norm_failed_attempts if isinstance(other, AiderTestResult) else self._normalize_attempts(other)
        self_val = self.norm_failed_attempts

        # If one is negative and the other isn't, the non-negative is "less than" (better)
        if (self_val < 0) != (other_val < 0):
            return self_val >= 0

        # Otherwise, normal comparison (works for both-negative and both-non-negative cases)
        return self_val < other_val

    # ATTENTION: __gt__ should not be required, but if not provided, the superclass method will be called,
    # causing incorrect comparisons.
    def __gt__(self, other: Union['AiderTestResult', int]) -> bool:
        if isinstance(other, (int, AiderTestResult)):
            return other < self
        return NotImplemented

def _get_attempt_limit_and_normalized_counts(benchmark_run: dict[str, AiderTestResult]) -> tuple[int | None, Counter]:
    result = Counter([t.failed_attempt_count for t in benchmark_run.values()])
    """
    Process and normalize the failed attempt counts from a benchmark run.

    Args:
    benchmark_run (dict[str, AiderTestResult]): Dictionary mapping test names to their results

    Returns:
    tuple[int | None, Counter]: A tuple containing:
    - The absolute value of the failure limit (if any tests hit it), or None
    - A Counter object containing:
    * Counts of tests with 0 or more failed attempts (these eventually passed)
    * Count of tests with -1 failed attempts (these never passed, hitting the failure limit)
    Note: All tests that hit the failure limit are normalized to count -1,
    regardless of the actual negative value used in the input
    """
    # Find the negative value (if any) and its count
    negative_value = next((k for k in result.keys() if k < 0), None)
    if negative_value is None:
        return 0, result
    # Get the count of tests with this negative value
    max_failed_attempts = result[negative_value]
    # Remove the original negative value from counter
    del result[negative_value]
    # Add the count to -1 in the counter
    result[-1] = max_failed_attempts
    return abs(negative_value), result


def parse_benchmark_source(benchmark_path: Traversable) -> list[AiderTestResult]:
    """
    Parse a benchmark run dir and extract test results.

    Args:
    benchmark_dir (Traversable): Path to the benchmark run dir.

    Returns:
    List[AiderTestResult]: A list of test results
    """
    with BenchmarkAnalyzer(benchmark_path) as analyzer:
        benchmark_results = analyzer.analyze(print_output=False)

    results = []
    for result_dict in benchmark_results.test_stats:
        # Convert the dictionary to match AiderTestResult format
        test_result = AiderTestResult(
            failed_attempt_count=result_dict['attempts'],
            word_count=result_dict['words'],
            name=result_dict['dir_name'],
            duration=result_dict['duration'],
            sent_tokens=result_dict['sent_tokens'],
            received_tokens=result_dict['received_tokens'],
            model=result_dict['model'],
            edit_format=result_dict['edit_format'],
            cost=result_dict['cost'],
            timeouts=result_dict['timeouts'],
            error_output_count=result_dict['error_outputs'],
            user_ask_count=result_dict['user_asks'],
            exhausted_context_window_count=result_dict['exhausted_context_windows'],
            malformed_responses=result_dict['malformed_responses'],
            syntax_errors=result_dict['syntax_errors'],
            indentation_errors=result_dict['indentation_errors'],
            lazy_comments=result_dict['lazy_comments'],
            cedarscript_errors=result_dict['cedarscript_errors']
        )
        results.append(test_result)

    return results


def compare_benchmark_runs(benchmark_run_1: dict[str, AiderTestResult], benchmark_run_2: dict[str, AiderTestResult]) -> tuple[list[str], list[str], list[str], list[str], list[str]]:
    """
    Compare two benchmark run dirs and categorize the changes.

    Args:
    benchmark_run_1 (dict[str, AiderTestResult]): First test result from first benchmark run, where keys are test names.
    benchmark_run_2 (dict[str, AiderTestResult]): Second test result from first benchmark run, in the same format as above.

    Returns:
    tuple[list[str], list[str], list[str], list[str], list[str]]: A tuple containing lists of:
        - tests only in benchmark_run_1
        - tests only in benchmark_run_2
        - improved tests
        - worsened tests
        - stable tests

    Tests are categorized based on their presence in the runs and changes in failed attempt counts.
    Negative failed run counts indicate the limit of failed attempts was reached and the test didn't pass.
    """
    only_1 = []
    only_2 = []
    improved = []
    worsened = []
    stable = []

    all_test_names = set(benchmark_run_1.keys()) | set(benchmark_run_2.keys())

    for test_name in sorted(all_test_names):
        test_from_run_1 = benchmark_run_1.get(test_name)
        test_from_run_2 = benchmark_run_2.get(test_name)

        if test_from_run_1 is None:
            only_2.append(test_name)
            continue
        if test_from_run_2 is None:
            only_1.append(test_name)
            continue
        if test_from_run_1 == test_from_run_2:
            stable.append(test_name)
            continue
        if test_from_run_2 < test_from_run_1:
            improved.append(test_name)
            continue
        worsened.append(test_name)

    return only_1, only_2, improved, worsened, stable


if __name__ == "__main__":
    load_dotenv(verbose=True)

    match len(sys.argv):
        case 3:
            benchmark1 = sys.argv[1]
            benchmark2 = sys.argv[2]
        case 2:
            benchmark1 = 'perfect'
            benchmark2 = sys.argv[1]
        case 1:
            benchmark1 = os.getenv('benchmark1')
            benchmark2 = os.getenv('benchmark2')
            if not benchmark2:
                benchmark2 = benchmark1
                benchmark1 = 'perfect'

    if not benchmark1 or not benchmark2:
        print(f"Usage: {sys.argv[0]} [benchmark-root-1] [benchmark-root-2]")
        print(f"Usage: benchmark1='<benchmark-root-1>' benchmark2='<benchmark-root-2>' {sys.argv[0]}")
        sys.exit(1)

    main(benchmark1, benchmark2)



#!/usr/bin/env python
from importlib.resources.abc import Traversable
from pathlib import Path
from collections import defaultdict

from dash import Dash, html, dcc, callback, Output, Input, dash_table
import plotly.graph_objects as go
import plotly.express as px
import statsmodels

import pandas as pd

import sys

if __name__ == "__main__":
    sys.path.append(str(Path(__file__).parent.parent))
from .delta_analyzer import AiderTestResult, parse_benchmark_source, compare_benchmark_runs
from .benchmark_analyzer import BenchmarkAnalyzer

app = Dash(__name__)

def create_dashboard(benchmark1: Traversable, benchmark2: Traversable, port: int = 8080):
    # Parse benchmark data
    benchmark_run_1 = {t.name: t for t in parse_benchmark_source(benchmark1)}
    benchmark_run_2 = {t.name: t for t in parse_benchmark_source(benchmark2)}

    # Get comparison results
    only_1, only_2, improved, worsened, stable = compare_benchmark_runs(
        benchmark_run_1, benchmark_run_2
    )

    # Create layout
    app.layout = html.Div([
        # Header
        html.H1('Benchmark Delta Dashboard', className='header'),

        # Benchmark Info
        html.Div([
            html.H2('Benchmark Information'),
            html.Div([
                html.Div([
                    html.H3('Baseline'),
                    html.P(f"Path: {benchmark1}"),
                    html.P(f"Tests: {len(benchmark_run_1)}"),
                ], className='info-box'),
                html.Div([
                    html.H3('Current'),
                    html.P(f"Path: {benchmark2}"),
                    html.P(f"Tests: {len(benchmark_run_2)}"),
                ], className='info-box'),
            ], className='info-container')
        ]),

        html.Div([
            html.H2('State Transition Analysis'),
            dcc.Graph(
                id='state-transition-sankey',
                figure=create_state_transition_sankey(benchmark_run_1, benchmark_run_2)
            ),
        ]),

        # Metrics Comparison
        html.Div([
            html.H2('Metrics Comparison'),
            dcc.Graph(
                id='metrics-comparison',
                figure=create_metrics_comparison(benchmark_run_1, benchmark_run_2)
            )
        ]),

        # Token Usage
        html.Div([
            html.H2('Token Usage'),
            create_token_usage_comparison(benchmark_run_1, benchmark_run_2)  # Use the returned Div directly
        ]),

        # TIOR vs. Errors
        html.Div([
            html.H2('TIOR vs. Errors'),
            dcc.Graph(
                id='tior-errors-scatter',
                figure=create_tior_vs_errors_scatter(benchmark_run_1, benchmark_run_2)
            )
        ]),

        # Token Ratio
        html.Div([
            html.H2('TIOR (Token I/O Ratio)'),
            dcc.Graph(
                id='token-ratio-bar-chart',
                figure=create_tior_bar_chart(benchmark_run_1, benchmark_run_2)
            )
        ]),

        # Detailed Test Results
        html.Div([
            html.H2('Detailed Test Results'),
            create_test_results_table(benchmark_run_1, benchmark_run_2)
        ]),
    ], className='container')

def create_metrics_comparison(run_1, run_2):
    """Create bar chart comparing various metrics"""

    metrics = {
        'Failed Tests': (sum(1 for t in run_1.values() if t.failed_attempt_count < 0),
                         sum(1 for t in run_2.values() if t.failed_attempt_count < 0)),
        'Failed Attempts': (sum(abs(t.failed_attempt_count) for t in run_1.values()),
                            sum(abs(t.failed_attempt_count) for t in run_2.values())),
        'Duration (minutes)': (sum(t.duration for t in run_1.values()) / 60,
                     sum(t.duration for t in run_2.values()) / 60),
        'Out Tokens (.5e-3)': (sum(t.received_tokens for t in run_1.values()) * .5e-3,
                     sum(t.received_tokens for t in run_2.values()) * .5e-3),
        'TIOR (1e-2)': (sum(t.sent_tokens / t.received_tokens for t in run_1.values() if t.received_tokens != 0) * 1e-2,
                     sum(t.sent_tokens / t.received_tokens for t in run_2.values() if t.received_tokens != 0) * 1e-2),
        'User Asks': (sum(t.user_ask_count for t in run_1.values()),
                      sum(t.user_ask_count for t in run_2.values())),
        'Errors (1e-1)': (sum(t.error_output_count for t in run_1.values()) * 1e-1,
                   sum(t.error_output_count for t in run_2.values()) * 1e-1),
        'CS Errors': (sum(t.cedarscript_errors for t in run_1.values()),
                      sum(t.cedarscript_errors for t in run_2.values())),
        'Syntax Errors': (sum(t.syntax_errors for t in run_1.values()),
                          sum(t.syntax_errors for t in run_2.values())),
        'Indentation Errors': (sum(t.indentation_errors for t in run_1.values()),
                          sum(t.indentation_errors for t in run_2.values())),
        'Timeouts': (sum(t.timeouts for t in run_1.values()),
                     sum(t.timeouts for t in run_2.values())),
    }

    fig = go.Figure(data=[
        go.Bar(name='Baseline', x=list(metrics.keys()), y=[m[0] for m in metrics.values()]),
        go.Bar(name='Current', x=list(metrics.keys()), y=[m[1] for m in metrics.values()])
    ])

    fig.update_layout(
        barmode='group',
        title_text="Key Metrics Comparison"
    )
    return fig

def create_tior_vs_errors_scatter(run_1, run_2):
    """Create a scatter plot with trend lines for TIOR vs. number of errors."""
    def calculate_tior(run):
        return {
            test_name: (test.sent_tokens / test.received_tokens if test.received_tokens != 0 else 0)
            for test_name, test in run.items()
        }

    tior_1 = calculate_tior(run_1)
    tior_2 = calculate_tior(run_2)

    df = pd.DataFrame({
        'Test': list(tior_1.keys()),
        'Baseline TIOR': list(tior_1.values()),
        'Current TIOR': [tior_2.get(test, 0) for test in tior_1.keys()],
        'Baseline Errors': [run_1[test].error_output_count for test in tior_1.keys()],
        'Current Errors': [run_2[test].error_output_count if test in run_2 else 0 for test in tior_1.keys()]
    })

    fig = px.scatter(
        df,
        x='Baseline TIOR',
        y='Baseline Errors',
        trendline='ols',
        title='TIOR vs. Errors (Baseline)',
        labels={'Baseline TIOR': 'Token Input/Output Ratio', 'Baseline Errors': 'Number of Errors'},
        hover_data=['Test']
    )

    fig.add_trace(
        px.scatter(
            df,
            x='Current TIOR',
            y='Current Errors',
            trendline='ols',
            title='TIOR vs. Errors (Current)',
            labels={'Current TIOR': 'Token Input/Output Ratio', 'Current Errors': 'Number of Errors'},
            hover_data=['Test']
        ).data[0]
    )

    fig.update_layout(
        title='TIOR vs. Errors with Trend Lines',
        xaxis_title='Token Input/Output Ratio',
        yaxis_title='Number of Errors',
        legend_title='Run'
    )

    return fig

def create_tior_bar_chart(run_1, run_2):
    """Create a bar chart for the ratio of sent_tokens to received_tokens."""
    def calculate_ratio(run):
        return {
            test_name: (test.sent_tokens / test.received_tokens if test.received_tokens != 0 else 0)
            for test_name, test in run.items()
        }

    ratio_1 = calculate_ratio(run_1)
    ratio_2 = calculate_ratio(run_2)

    df = pd.DataFrame({
        'Test': list(ratio_1.keys()),
        'Baseline Ratio': list(ratio_1.values()),
        'Current Ratio': [ratio_2.get(test, 0) for test in ratio_1.keys()]
    })

    fig = go.Figure(data=[
        go.Bar(name='Baseline', x=df['Test'], y=df['Baseline Ratio']),
        go.Bar(name='Current', x=df['Test'], y=df['Current Ratio'])
    ])

    fig.update_layout(
        barmode='group',
        title_text="TIOR (Token I/O Ratio)"
    )
    return fig


def create_token_usage_comparison(run_1, run_2):
    """Create two scatter plots comparing token usage for baseline and current runs"""
    # Create dataframes with error handling for missing values
    def create_df(run):
        return pd.DataFrame({
            'Test': list(run.keys()),
            'input': [t.sent_tokens for t in run.values()],
            'output': [t.received_tokens for t in run.values()],
            'Total Tokens': [t.sent_tokens + t.received_tokens for t in run.values()],
            'Errors': [t.error_output_count for t in run.values()]
        }).fillna(0)  # Handle any missing values

    df_baseline = create_df(run_1)
    df_current = create_df(run_2)

    # Get max values for consistent scaling
    max_input = max(df_baseline['input'].max(), df_current['input'].max())
    max_output = max(df_baseline['output'].max(), df_current['output'].max())
    max_total = max(df_baseline['Total Tokens'].max(), df_current['Total Tokens'].max())
    max_errors = max(df_baseline['Errors'].max(), df_current['Errors'].max())

    def create_scatter(df, title):
        return px.scatter(
            df,
            x='input',
            y='output',
            size='Total Tokens',
            size_max=30,  # Limit maximum bubble size
            color='Errors',
            range_color=[0, max_errors],  # Consistent color scale
            hover_data=['Test'],
            title=title,
            labels={
                'input': 'Input Tokens',
                'output': 'Output Tokens'
            }
        )

    # Create plots with consistent scales
    fig_baseline = create_scatter(df_baseline, 'Baseline Token Usage')
    fig_current = create_scatter(df_current, 'Current Token Usage')

    # Update layout for both plots
    for fig in [fig_baseline, fig_current]:
        fig.update_layout(
            showlegend=True,
            plot_bgcolor='white',
            width=1200,
            height=400,
            margin=dict(t=50, b=50, l=50, r=50),
            xaxis=dict(range=[0, max_input * 1.1]),  # Add 10% padding
            yaxis=dict(range=[0, max_output * 1.1])
        )
        fig.update_traces(
            marker=dict(line=dict(width=1, color='DarkSlateGrey')),
            selector=dict(mode='markers')
        )
        fig.update_xaxes(showgrid=True, gridwidth=1, gridcolor='LightGrey')
        fig.update_yaxes(showgrid=True, gridwidth=1, gridcolor='LightGrey')

    return html.Div(children=[
        html.Div(children=[
            dcc.Graph(
                id='token-usage-baseline',
                figure=fig_baseline
            )
        ], className='six columns'),
        html.Div(children=[
            dcc.Graph(
                id='token-usage-current',
                figure=fig_current
            )
        ], className='six columns')
    ], className='row')


def create_token_usage_comparison_all(run_1, run_2):
    """Create scatter plot comparing token usage"""
    df = pd.DataFrame({
        'Test': list(run_1.keys()),
        'Sent (Baseline)': [t.sent_tokens for t in run_1.values()],
        'Received (Baseline)': [t.received_tokens for t in run_1.values()],
        'Sent (Current)': [run_2[k].sent_tokens if k in run_2 else 0 for k in run_1.keys()],
        'Received (Current)': [run_2[k].received_tokens if k in run_2 else 0 for k in run_1.keys()],
    })

    fig = px.scatter(
        df,
        x='Sent (Baseline)',
        y='Received (Baseline)',
        size='Sent (Current)',
        color='Received (Current)',
        hover_data=['Test']
    )

    fig.update_layout(title_text="Token Usage Distribution")
    return fig

def create_test_results_table(run_1, run_2):
    """Create an interactive table showing detailed test results with token ratios."""
    data = []
    for test in sorted(set(run_1.keys()) | set(run_2.keys())):
        tior_baseline = (run_1[test].sent_tokens / run_1[test].received_tokens
                          if test in run_1 and run_1[test].received_tokens != 0 else 0)
        tior_current = (run_2[test].sent_tokens / run_2[test].received_tokens
                         if test in run_2 and run_2[test].received_tokens != 0 else 0)
        tior_percent_change = ((tior_current - tior_baseline) / tior_baseline * 100
                          if tior_baseline != 0 else 'N/A')

        if test in run_1 and test in run_2:
            data.append({
                'Test Name': test,
                'Status': get_test_status(test, run_1, run_2),
                'Baseline Failed Attempts': f"{run_1[test].failed_attempt_count:+d}",
                'Current Failed Attempts': f"{run_2[test].failed_attempt_count:+d}",
                'Δ Duration': f"{run_2[test].duration - run_1[test].duration:+.2f}s",
                'Δ Out Token': f"{run_2[test].received_tokens - run_1[test].received_tokens:+d}",
                'Δ CS Errors': f"{run_2[test].cedarscript_errors - run_1[test].cedarscript_errors:+d}",
                'Δ Syntax Errors': f"{run_2[test].syntax_errors - run_1[test].syntax_errors:+d}",
                'Δ Errors': f"{run_2[test].error_output_count - run_1[test].error_output_count:+d}",
                'Δ TIOR': f"{tior_percent_change:.2f}%" if tior_percent_change != 'N/A' else 'N/A',
                'Baseline TIOR': f"{tior_baseline:.2f}",
                'Current TIOR': f"{tior_current:.2f}"
            })
        else:
            data.append({
                'Test Name': test,
                'Status': get_test_status(test, run_1, run_2),
                'Baseline Failed Attempts': 'N/A' if test not in run_1 else f"{run_1[test].failed_attempt_count:.2f}",
                'Current Failed Attempts': 'N/A' if test not in run_2 else f"{run_2[test].failed_attempt_count:.2f}",
                'Δ Duration': 'N/A',
                'Δ Out Token': 'N/A',
                'Δ CS Errors': 'N/A',
                'Δ Syntax Errors': 'N/A',
                'Δ Errors': 'N/A',
                'Δ TIOR': 'N/A',
                'Baseline TIOR': 'N/A' if test not in run_1 else f"{tior_baseline:.2f}",
                'Current TIOR': 'N/A' if test not in run_2 else f"{tior_current:.2f}"
            })

    return dash_table.DataTable(
        id='results-table',
        columns=[{'name': k, 'id': k} for k in data[0].keys()],
        data=data,
        style_table={'overflowX': 'auto'},
        style_cell={
            'textAlign': 'left',
            'padding': '12px',
            'whiteSpace': 'normal',
            'height': 'auto',
        },
        style_header={
            'backgroundColor': 'rgb(230, 230, 230)',
            'fontWeight': 'bold'
        },
        style_data_conditional=[
            {
                'if': {'row_index': 'odd'},
                'backgroundColor': 'rgb(248, 248, 248)'
            }
        ],
        filter_action="native",
        sort_action="native",
        sort_mode="multi",
        page_size=20
    )

def get_test_status(test, run_1, run_2):
    """Determine the status of a test between runs"""
    if test not in run_1:
        return "New"
    if test not in run_2:
        return "Missing"
    if run_1[test] == run_2[test]:
        return "Stable: OK" if run_1[test].failed_attempt_count >= 0 else "Stable: Failed"
    if run_1[test] > run_2[test]:
        return "Now Passed" if run_1[test].failed_attempt_count < 0 else "Minor Improvement"
    return "Now Failed" if run_2[test].failed_attempt_count < 0 else "Degraded"

def create_state_transition_sankey(run_1: dict, run_2: dict):
    # Helper function to get state
    def get_state(aider_test_result: AiderTestResult):
        if aider_test_result is None:
            return "Missing"
        if aider_test_result.norm_failed_attempts < 0:
            return "Failed"
        if aider_test_result.norm_failed_attempts == 0:
            return "Passed on first"
        return f"Passed after {aider_test_result.norm_failed_attempts}"

    # Define states and their colors
    states = ["Failed", "Missing", "Passed on first"] + [f"Passed after {i}" for i in range(1, 2)]
    colors = {
        "Failed": "#ff4444",
        "Missing": "#aaaaaa",
        "Passed on first": "#00cc00",
        "Passed after 1": "#F3E400",
        "Passed after 2": "#FFAC1E",
    }

    # Create node labels
    left_nodes = [f"Baseline: {state}" for state in states]
    right_nodes = [f"Current: {state}" for state in states]
    all_nodes = left_nodes + right_nodes

    # Create node index mapping
    node_mapping = {name: idx for idx, name in enumerate(all_nodes)}

    # Initialize flow tracking
    flows = defaultdict(int)

    # Calculate transitions for all unique test names
    for test in set(run_1.keys()) | set(run_2.keys()):
        state1 = get_state(run_1.get(test, None))
        state2 = get_state(run_2.get(test, None))

        flows[(f"Baseline: {state1}", f"Current: {state2}")] += 1

    # Prepare Sankey data
    source = []
    target = []
    value = []

    for (src, tgt), count in flows.items():
        if count > 0:  # Only add non-zero flows
            source.append(node_mapping[src])
            target.append(node_mapping[tgt])
            value.append(count)

    # Create node colors list
    node_colors = []
    for label in all_nodes:
        state = label.split(": ")[1]
        node_colors.append(colors[state])

    # Create the figure
    fig = go.Figure(data=[go.Sankey(
        node=dict(
            pad=15,
            thickness=20,
            line=dict(color="black", width=0.5),
            label=all_nodes,
            color=node_colors
        ),
        link=dict(
            source=source,
            target=target,
            value=value,
            color=[f"rgba(169, 169, 169, 0.3)"] * len(source)  # Semi-transparent gray links
        )
    )])

    # Update layout
    fig.update_layout(
        title_text="State Transitions Between Runs",
        font_size=10,
        height=600,
        margin=dict(t=25, l=25, r=25, b=25)
    )

    return fig

if __name__ == '__main__':
    import sys
    from dotenv import load_dotenv
    import os

    load_dotenv(verbose=True)

    match len(sys.argv):
        case 3:
            benchmark1 = Path(sys.argv[1])
            benchmark2 = Path(sys.argv[2])
        case 2:
            benchmark1 = 'perfect'
            benchmark2 = Path(sys.argv[1])
        case 1:
            benchmark1 = Path(os.getenv('benchmark1'))
            benchmark2 = os.getenv('benchmark2')
            if not benchmark2:
                benchmark2 = benchmark1
                benchmark1 = 'perfect'
            else:
                benchmark2 = Path(benchmark2)

    if not benchmark1 or not benchmark2:
        print(f"Usage: bda [benchmark-root-1] [benchmark-root-2]")
        print(f"Usage: benchmark1='<benchmark-root-1>' benchmark2='<benchmark-root-2>' bda")
        sys.exit(1)

    create_dashboard(benchmark1, benchmark2)
    app.run(debug=True)



from pathlib import Path
from importlib.resources.abc import Traversable


def find_aider_results(path_or_traversable: Path | Traversable):
    return find_files(path_or_traversable, '.aider.results.json')


def find_files(path_or_traversable: Path | Traversable, file_pattern: str) -> list:
    if isinstance(path_or_traversable, Path):
        return sorted(path_or_traversable.rglob(file_pattern))

    # For Traversable implementation
    def _recursive_search(traversable: Traversable) -> Iterator[Traversable]:
        if traversable.is_file() and traversable.name == file_pattern:
            yield traversable
        elif traversable.is_dir():
            for child in traversable.iterdir():
                yield from _recursive_search(child)

    return sorted(_recursive_search(path_or_traversable), key=lambda x: str(x))



from importlib.resources.abc import Traversable

from version import __version__
import argparse
import sys
import os
from pathlib import Path

def show_dashboard(run1: Traversable, run2: Traversable, port: int = 8050):
    from bda import app, create_dashboard
    create_dashboard(run1, run2, port=port)
    app.run(debug=True)

def show_info(run: Path):
    from bda import show_info
    show_info(run)

def show_delta(run1: Path, run2: Path):
    from bda import show_delta
    show_delta(run1, run2)

def create_parser():
    parser = argparse.ArgumentParser(
        description="Analyze and compare Aider benchmark runs"
    )

    # Create mutually exclusive group for modes
    mode_group = parser.add_mutually_exclusive_group()
    mode_group.add_argument(
        "--dashboard",
        action="store_true",
        help="Launch interactive web dashboard"
    )
    mode_group.add_argument(
        "--info",
        action="store_true",
        help="Show information about a single run"
    )

    # Optional benchmark arguments - we'll handle defaults ourselves
    parser.add_argument(
        "benchmarks",
        nargs="*",
        type=str,
        help="Benchmark run(s) to analyze. Can point to a folder or a tar-bzip2 archive; Can be specified via environment variables benchmark1 and benchmark2"
    )

    parser.add_argument(
        "--port",
        type=int,
        default=8050,
        help="Port for dashboard web server (default: 8050)"
    )

    return parser

def main():
    parser = create_parser()
    from util.cli_kit import get_benchmark_args
    mode, benchmark1, benchmark2, port = get_benchmark_args(parser)

    if not benchmark1 or not benchmark2 and mode != 'info':
        parser.error(
            "Benchmarks must be specified either as arguments or via environment variables:\n"
            f"  bda [benchmark-root-1] [benchmark-root-2]\n"
            f"  benchmark1='<benchmark-root-1>' benchmark2='<benchmark-root-2>' bda"
        )

    match mode:
        case 'info':
            show_info(Path(benchmark1))
        case 'dashboard':
            show_dashboard(Path(benchmark1), Path(benchmark2), port)
        case _:  # delta mode
            show_delta(Path(benchmark1), Path(benchmark2))

    return 0

if __name__ == "__main__":
    sys.exit(main())



from dotenv import load_dotenv
import os

def get_benchmark_args(parser):
    load_dotenv(verbose=True)
    args = parser.parse_args()

    if args.info:
        if len(args.benchmarks) != 1:
            parser.error("--info requires exactly one benchmark")
        return 'info', args.benchmarks[0], None

    # Handle benchmark arguments
    match len(args.benchmarks):
        case 2:
            b1, b2 = args.benchmarks
        case 1:
            b1, b2 = 'perfect', args.benchmarks[0]
        case 0:
            b1 = os.getenv('benchmark1')
            b2 = os.getenv('benchmark2')
            if not b2:
                b2 = b1
                b1 = 'perfect'
        case _:
            parser.error("Too many benchmark arguments")

    mode = 'dashboard' if args.dashboard else 'delta'
    return mode, b1, b2, args.port




#!/usr/bin/env python
import json
import os
import shutil
import sys
from datetime import datetime, timedelta
import re
from importlib.resources.abc import Traversable
from os import getenv
from pathlib import Path
from dataclasses import dataclass

from bda.file_kit import find_files, find_aider_results
from dotenv import load_dotenv
import tarfile
import tempfile


@dataclass
class BenchmarkResult:
    duration: float
    success_rate: float
    total_tests: int
    passed_tests: int
    failed_tests: int
    failed_attempts: int
    total_sent_tokens: int
    total_received_tokens: int
    total_cost: float
    total_test_timeouts: int
    total_num_error_outputs: int
    total_num_user_asks: int
    total_num_exhausted_context_windows: int
    total_num_malformed_responses: int
    total_syntax_errors: int
    total_indentation_errors: int
    total_lazy_comments: int
    total_cedarscript_errors: int
    total_word_count: int
    test_stats: list[dict]


def archive2dir(archive_path: Path) -> Path:
    with tempfile.TemporaryDirectory(prefix="benchmark-", delete=False) as temp_dir:
        try:
            with tarfile.open(archive_path, 'r:bz2') as tar:
                tar.extractall(path=temp_dir)
            # Update benchmark_run_path to point to extracted contents
            extracted_contents = list(Path(temp_dir).iterdir())
            if len(extracted_contents) == 1 and extracted_contents[0].is_dir():
                return  extracted_contents[0]
            return Path(temp_dir)
        except Exception as e:
            shutil.rmtree(temp_dir, ignore_errors=True)
            raise RuntimeError(f"Failed to extract archive: {e}")


class BenchmarkAnalyzer:
    def __init__(self, benchmark_run_path: Traversable):
        load_dotenv(verbose=True)
        self.must_delete_benchmark_dir = False
        self.is_perfect_run = str(benchmark_run_path) == 'perfect'
        if self.is_perfect_run:
            return

        self.benchmark_run_path = benchmark_run_path
        match self.benchmark_run_path:
            case Path() as path if not path.is_absolute():
                benchmark_root = getenv('benchmark_root')
            case _: benchmark_root = None

        if benchmark_root:
            benchmark_root = Path(benchmark_root)
            if not benchmark_root.is_absolute():
                benchmark_root = Path.cwd() / benchmark_root

            self.benchmark_run_path = benchmark_root / self.benchmark_run_path

        if isinstance(self.benchmark_run_path, Path) and not self.benchmark_run_path.exists():
            raise FileNotFoundError(f"Benchmark not found: {self.benchmark_run_path}")

    def __enter__(self) -> 'BenchmarkAnalyzer':
        # Support usage as a context manager
        # Determine if the path is a directory or an archive
        if not self.benchmark_run_path.is_dir():
            # Assume it's a tar.bz2 archive
            self.benchmark_run_path = archive2dir(self.benchmark_run_path)
            self.must_delete_benchmark_dir = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.cleanup()

    def __del__(self) -> None:
        self.cleanup()

    def cleanup(self) -> None:
        """Cleanup temporary directory if it exists"""
        if self.must_delete_benchmark_dir and  self.benchmark_run_path:
            shutil.rmtree(self.benchmark_run_path, ignore_errors=True)

    @staticmethod
    def create_perfect_result() -> BenchmarkResult:
        """Create an idealized perfect benchmark result."""
        return BenchmarkResult(
            duration=1.0,
            success_rate=100.0,
            total_tests=1,  # At least one test to avoid division by zero
            passed_tests=1,
            failed_tests=0,
            failed_attempts=0,
            total_sent_tokens=1,
            total_received_tokens=1,
            total_cost=0.0,
            total_test_timeouts=0,
            total_num_error_outputs=0,
            total_num_user_asks=0,
            total_num_exhausted_context_windows=0,
            total_num_malformed_responses=0,
            total_syntax_errors=0,
            total_indentation_errors=0,
            total_lazy_comments=0,
            total_cedarscript_errors=0,
            total_word_count=0,
            test_stats=[{
                'attempts': 0,
                'dir_name': 'perfect',
                'duration': 1.0,
                'sent_tokens': 1,
                'received_tokens': 1,
                'model': 'perfect',
                'edit_format': 'perfect',
                'cost': 0.0,
                'timeouts': 0,
                'error_outputs': 0,
                'user_asks': 0,
                'exhausted_context_windows': 0,
                'malformed_responses': 0,
                'syntax_errors': 0,
                'indentation_errors': 0,
                'lazy_comments': 0,
                'cedarscript_errors': 0,
                'words': 0
            }]
        )

    @staticmethod
    def aider_results_to_cols(json_file: Path) -> tuple:
        with open(json_file, 'r', encoding='utf-8') as f:
            data = json.load(f)
        return (data['model'], data['edit_format'], data['cost'],
                data['duration'], data['test_timeouts'], data['num_error_outputs'],
                data['num_user_asks'], data['num_exhausted_context_windows'],
                data['num_malformed_responses'], data['syntax_errors'],
                data['indentation_errors'], data['lazy_comments'])

    @staticmethod
    def format_duration(seconds: float) -> str:
        return str(timedelta(seconds=int(seconds)))

    @staticmethod
    def extract_word_count(root: Path) -> int:
        whitespace_pattern = re.compile(r'\s+')
        word_count = 0
        chunk_size = 1024 * 1024  # 1MB chunks

        files = [f for f in root.glob("*.py") if not f.name.endswith("_test.py")]

        for file_path in files:
            with open(file_path, 'r', encoding='utf-8', errors='ignore') as f:
                # Process file in chunks for memory efficiency
                while chunk := f.read(chunk_size):
                    # Process non-empty lines
                    lines = filter(None, (line.strip() for line in chunk.splitlines()))

                    # Count words using pre-compiled regex
                    word_count += sum(len(whitespace_pattern.split(line)) for line in lines)

        return word_count

    @staticmethod
    def extract_chat_history_counts(file_path: Path) -> tuple[int, int, int]:
        total_sent = total_received = cedarscript_error_count = 0

        with open(file_path, 'r', encoding='utf-8') as f:
            for line in f:
                if '> Tokens:' in line:
                    parts = line.split()
                    for i, part in enumerate(parts):
                        if part.lower().startswith(('sent', 'sent,', 'sent.')):
                            sent = parts[i-1]
                            multiplier = 1000 if 'k' in sent.lower() else 1000000 if 'm' in sent.lower() else 1
                            sent = sent.lower().replace('k', '').replace('m', '')
                            try:
                                total_sent += float(sent) * multiplier
                            except ValueError:
                                continue

                        if part.lower().startswith(('received', 'received,', 'received.')):
                            received = parts[i-1]
                            multiplier = 1000 if 'k' in received.lower() else 1000000 if 'm' in received.lower() else 1
                            received = received.lower().replace('k', '').replace('m', '')
                            try:
                                total_received += float(received) * multiplier
                            except ValueError:
                                continue

                if '<error-location>COMMAND #' in line:
                    cedarscript_error_count += 1

        return int(total_sent), int(total_received), cedarscript_error_count

    def analyze(self, print_output: bool = False) -> BenchmarkResult | None:
        if self.is_perfect_run:
            result = self.create_perfect_result()
            if print_output:
                self._print_results(result)
            return result

        i = total_word_count = total_duration = total_failed_attempts = failed_test_count = 0
        total_sent_tokens = total_received_tokens = total_cost = 0
        total_test_timeouts = total_num_error_outputs = total_num_user_asks = 0
        total_num_exhausted_context_windows = total_num_malformed_responses = 0
        total_syntax_errors = total_indentation_errors = total_lazy_comments = 0
        total_cedarscript_errors = 0
        detailed_results = []

        if print_output:
            print(f"# -dirname {self.benchmark_run_path.name} tests")
            print('failed-attempts (negative if all attempts failed), words, test-name, duration, sent_tokens, received_tokens, '
                  'model, edit_format, cost, test_timeouts, num_error_outputs, num_user_asks, num_exhausted_context_windows, '
                  'num_malformed_responses, syntax_errors, indentation_errors, lazy_comments, cedarscript_errors')

        for aider_json_file in find_aider_results(self.benchmark_run_path):
            i += 1
            test_folder: Path = aider_json_file.parent
            chat_history_file: Path = (test_folder / '.aider.chat.history.md').absolute()
            if not chat_history_file.exists():
                if test_folder.exists():
                    print(f"[analyze] File '{chat_history_file.name}' not found in {test_folder}")
                else:
                    print(f"[analyze] Test folder '{test_folder.name}' not found in {test_folder.parent}")
                return None
            word_count = self.extract_word_count(test_folder)
            total_word_count += word_count
            sent_tokens, received_tokens, test_cedarscript_errors = self.extract_chat_history_counts(chat_history_file)
            total_sent_tokens += sent_tokens
            total_received_tokens += received_tokens
            total_cedarscript_errors += test_cedarscript_errors

            with open(aider_json_file, 'r', encoding='utf-8') as f:
                data = json.load(f)
                outcome = data['tests_outcomes']

            attempts = 0 if outcome == [True] else outcome.count(False)
            total_failed_attempts += attempts

            dir_name = test_folder.name
            if True not in outcome:
                attempts = -attempts
                failed_test_count += 1

            test_results = self.aider_results_to_cols(aider_json_file)
            (test_model, test_edit_format, test_cost, test_duration, test_timeouts,
             test_num_error_outputs, test_num_user_asks, test_num_exhausted_context_windows,
             test_num_malformed_responses, test_syntax_errors, test_indentation_errors,
             test_lazy_comments) = test_results

            # Update totals
            total_duration += float(test_duration)
            total_cost += float(test_cost)
            total_test_timeouts += test_timeouts
            total_num_error_outputs += test_num_error_outputs
            total_num_user_asks += test_num_user_asks
            total_num_exhausted_context_windows += test_num_exhausted_context_windows
            total_num_malformed_responses += test_num_malformed_responses
            total_syntax_errors += test_syntax_errors
            total_indentation_errors += test_indentation_errors
            total_lazy_comments += test_lazy_comments

            result_dict = {
                'attempts': attempts,
                'dir_name': dir_name,
                'duration': test_duration,
                'sent_tokens': sent_tokens,
                'received_tokens': received_tokens,
                'model': test_model,
                'edit_format': test_edit_format,
                'cost': test_cost,
                'timeouts': test_timeouts,
                'error_outputs': test_num_error_outputs,
                'user_asks': test_num_user_asks,
                'exhausted_context_windows': test_num_exhausted_context_windows,
                'malformed_responses': test_num_malformed_responses,
                'syntax_errors': test_syntax_errors,
                'indentation_errors': test_indentation_errors,
                'lazy_comments': test_lazy_comments,
                'cedarscript_errors': test_cedarscript_errors,
                'words': word_count
            }
            detailed_results.append(result_dict)

        success_rate = 100 * (i-failed_test_count) / i if i > 0 else 0

        result = BenchmarkResult(
            duration=total_duration,
            success_rate=success_rate,
            total_tests=i,
            passed_tests=i-failed_test_count,
            failed_tests=failed_test_count,
            failed_attempts=total_failed_attempts,
            total_sent_tokens=total_sent_tokens,
            total_received_tokens=total_received_tokens,
            total_cost=total_cost,
            total_test_timeouts=total_test_timeouts,
            total_num_error_outputs=total_num_error_outputs,
            total_num_user_asks=total_num_user_asks,
            total_num_exhausted_context_windows=total_num_exhausted_context_windows,
            total_num_malformed_responses=total_num_malformed_responses,
            total_syntax_errors=total_syntax_errors,
            total_indentation_errors=total_indentation_errors,
            total_lazy_comments=total_lazy_comments,
            total_cedarscript_errors=total_cedarscript_errors,
            total_word_count=total_word_count,
            test_stats=detailed_results
        )

        if print_output:
            self._print_results(result)

        return result

    def _print_results(self, result: BenchmarkResult):
        """Print formatted benchmark results."""
        print(f"# -dirname {self.benchmark_run_path.name if not self.is_perfect_run else 'perfect'} tests")
        print('failed-attempts (negative if all attempts failed), words, test-name, duration, sent_tokens, received_tokens, '
              'model, edit_format, cost, test_timeouts, num_error_outputs, num_user_asks, num_exhausted_context_windows, '
              'num_malformed_responses, syntax_errors, indentation_errors, lazy_comments, cedarscript_errors')

        for test in result.test_stats:
            print(
                f'{test["attempts"]:2d}, {test["words"]:5d}, {test["dir_name"]:25s}, {test["duration"]:7.3f}, '
                f'{test["sent_tokens"]:6d}, {test["received_tokens"]:07d}, '
                f'{test["model"]:25s}, {test["edit_format"]:25s}, {test["cost"]:0.3f}, '
                f'{test["timeouts"]:2d}, {test["error_outputs"]:2d}, {test["user_asks"]:2d}, '
                f'{test["exhausted_context_windows"]:2d}, {test["malformed_responses"]:2d}, '
                f'{test["syntax_errors"]:2d}, {test["indentation_errors"]:2d}, '
                f'{test["lazy_comments"]:2d}, {test["cedarscript_errors"]:2d}'
            )

        print('='*17)
        print(f'Words        : {result.total_word_count}')
        print(f'Duration     : {self.format_duration(result.duration)}')
        print(f'Success      : {result.success_rate:03.1f}% ( {result.passed_tests} / {result.total_tests} )')
        print('# duration_s, test_pass_count, test_failed_count, failed_attempts, total_sent_tokens, '
              'total_received_tokens, total_cost, total_test_timeouts, total_num_error_outputs, '
              'total_num_user_asks, total_num_exhausted_context_windows, total_num_malformed_responses, '
              'total_syntax_errors, total_indentation_errors, total_lazy_comments, total_cedarscript_errors, total_words')
        print(
            f'# {result.duration:03.3f}, {result.passed_tests:03d}, {result.failed_tests:03d}, '
            f'{result.failed_attempts:03d}, {result.total_sent_tokens:07d}, {result.total_received_tokens:07d}, '
            f'{result.total_cost:0.3f}, {result.total_test_timeouts:02d}, {result.total_num_error_outputs:02d}, '
            f'{result.total_num_user_asks:02d}, {result.total_num_exhausted_context_windows:02d}, '
            f'{result.total_num_malformed_responses:02d}, {result.total_syntax_errors:02d}, '
            f'{result.total_indentation_errors:02d}, {result.total_lazy_comments:02d}, '
            f'{result.total_cedarscript_errors:02d}, {result.total_word_count:02d}'
        )


def main():
    load_dotenv(verbose=True)
    match len(sys.argv):
        case 2:
            benchmark_path = sys.argv[1]
        case _:
            benchmark_path = getenv('benchmark_path')
    if not benchmark_path:
        print(f"Usage: {sys.argv[0]} [benchmark-path]")
        print(f"Usage: benchmark_path='<benchmark-path>' {sys.argv[0]}")
        sys.exit(1)

    with BenchmarkAnalyzer(benchmark_path) as analyzer:
        analyzer.analyze(print_output=True)


if __name__ == '__main__':
    main()



from version import __version__

from .dashboard import app, create_dashboard
from .benchmark_analyzer import BenchmarkAnalyzer
from .delta_analyzer import main as show_delta


__all__ = [
    "__version__", "create_dashboard", "show_info", "show_delta"
]

def show_info(benchmark_path: str):
    with BenchmarkAnalyzer(benchmark_path) as analyzer:
        analyzer.analyze(print_output=True)
